import logging

from torch import exp, nn, tensor
import networkx as nx
import torch
import torch.nn.functional as F

from chebai.models.base import JCIBaseNet

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class Recursive(JCIBaseNet):
    NAME = "REC"

    def __init__(self, in_d, out_d, num_classes, **kwargs):
        super().__init__(num_classes, **kwargs)
        mem_len = in_d
        self.internal_dimension = in_d
        self.embedding = nn.Embedding(800, 100)

        self.input_post = nn.Linear(in_d, in_d)

        self.input_attention = nn.MultiheadAttention(in_d, 5)
        self.hidden_attention = nn.MultiheadAttention(in_d, 5)
        self.merge_attention = nn.MultiheadAttention(in_d, 5)

        self.hidden_post = nn.Linear(in_d, in_d)

        self.merge_post = nn.Linear(in_d, in_d)

        self.post = nn.Linear(in_d, in_d)

        self.children_attention = nn.MultiheadAttention(in_d, 5)

        self.input_norm_1 = nn.LayerNorm(in_d)
        self.input_norm_2 = nn.LayerNorm(in_d)
        self.hidden_norm_1 = nn.LayerNorm(in_d)
        self.merge_norm_1 = nn.LayerNorm(in_d)
        self.merge_norm_2 = nn.LayerNorm(in_d)

        self.base = torch.nn.parameter.Parameter(torch.empty((in_d,)))
        self.base_memory = torch.nn.parameter.Parameter(torch.empty((mem_len,)))
        self.output = nn.Sequential(
            nn.Linear(in_d, in_d),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(in_d, num_classes),
        )

    def forward(self, batch):
        result = []
        for row in batch:
            graph = row[0]
            c = nx.center(graph)[0]
            d = nx.single_source_shortest_path(graph, c)
            if graph.edges:
                digraph = nx.DiGraph(
                    (a, b) if d[a] > d[b] else (b, a) for (a, b) in graph.edges
                )
            else:
                digraph = nx.DiGraph(graph)
            child_results = {}
            x = None
            for node in nx.topological_sort(digraph):
                child_values = child_results.pop(node, [])
                inp = self.embedding(graph.nodes[node]["x"])
                if not child_values:
                    hidden_state = self.base_memory
                else:
                    hidden_state = self.merge_childen(child_values, inp)
                x = self.input(inp, hidden_state)
                for s in digraph.successors(node):
                    child_results[s] = child_results.get(s, []) + [x]
            result.append(self.output(x))
        return torch.stack(result)

    def merge_childen(self, child_values, x):
        stack = torch.stack(child_values).unsqueeze(0).transpose(1, 0)
        att = self.children_attention(
            x.expand(1, stack.shape[1], -1).transpose(1, 0), stack, stack
        )[0]
        return torch.sum(att.squeeze(0), dim=0)

    def input(self, x0, hidden):

        x = x0.unsqueeze(0).unsqueeze(0)
        a = self.input_norm_1(x + self.input_attention(x, x, x)[0])
        a = self.input_norm_2(a + F.relu(self.input_post(a)))

        h0 = hidden.unsqueeze(0).unsqueeze(0)
        b = self.hidden_norm_1(h0 + self.input_attention(h0, h0, h0)[0])
        # b = self.norm(b + self.hidden_post(b))

        c = self.merge_norm_1(b + self.merge_attention(a, b, b)[0])
        c = self.merge_norm_2(c + F.relu(self.merge_post(c)))

        return self.post(c).squeeze(0).squeeze(0)



import itertools
import logging
import os
import sys

from pytorch_lightning import loggers as pl_loggers
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from pytorch_lightning.tuner.tuning import Tuner
from sklearn.metrics import f1_score
from torch import nn
from torchmetrics import MeanSquaredError
from torchmetrics import F1Score, MeanSquaredError
from torchmetrics import functional as tmf
import pytorch_lightning as pl
import torch
import torchmetrics
import tqdm

from chebai.preprocessing.datasets import XYBaseDataModule

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class JCIBaseNet(pl.LightningModule):
    NAME = None
    LOSS = torch.nn.BCEWithLogitsLoss

    def __init__(self, loss_cls=None, out_dim=None, **kwargs):
        super().__init__()
        self.save_hyperparameters()
        if out_dim and out_dim > 1:
            task = "multilabel"
        else:
            task = "binary"
        self.out_dim = out_dim
        weights = kwargs.get("weights", None)
        if loss_cls is None:
            loss_cls = torch.nn.BCEWithLogitsLoss
        if weights is not None:
            self.loss = loss_cls(pos_weight=weights)
        else:
            self.loss = loss_cls()

        self.optimizer_kwargs = kwargs.get("optimizer_kwargs", dict())
        self.thres = kwargs.get("threshold", 0.5)
        self.metrics = ["F1Score", "Precision", "Recall", "AUROC"]
        self.metric_aggs = ["micro"]
        for metric in self.metrics:
            for agg in self.metric_aggs:
                setattr(
                    self,
                    metric + agg,
                    getattr(torchmetrics, metric)(
                        threshold=self.thres,
                        average=agg,
                        task=task,
                        num_labels=self.out_dim,
                    ),
                )

    def _get_prediction_and_labels(self, data, labels, output):
        return output, labels

    def _get_data_and_labels(self, batch, batch_idx):
        return dict(features=batch.x, labels=batch.y.float())

    def _execute(self, batch, batch_idx):
        data = self._get_data_and_labels(batch, batch_idx)
        labels = data["labels"]
        model_output = self(data, **data.get("model_kwargs", dict()))
        return data, labels, model_output

    def _get_data_for_loss(self, model_output, labels):
        return dict(input=model_output, target=labels.float())

    def training_step(self, *args, **kwargs):
        return self.calculate_all_metrics("train_", *args, on_step=True, **kwargs)

    def validation_step(self, *args, **kwargs):
        with torch.no_grad():
            return self.calculate_all_metrics("val_", *args, **kwargs)

    def test_step(self, *args, **kwargs):
        with torch.no_grad():
            return self.calculate_all_metrics("test_", *args, **kwargs)

    def calculate_all_metrics(self, prefix, *args, on_step=False, **kwargs):
        data, labels, model_output = self._execute(*args, **kwargs)
        loss = self.loss(**self._get_data_for_loss(model_output, labels))
        with torch.no_grad():
            p, l = self._get_prediction_and_labels(data, labels, model_output)
            for name in self.metrics:
                for agg in self.metric_aggs:
                    metric = getattr(self, name + agg)
                    self.log(
                        prefix + name + "_" + agg,
                        metric(preds=p.detach().cpu(), target=l.detach().cpu()),
                        on_step=on_step,
                        on_epoch=True,
                        prog_bar=True,
                        logger=True,
                        batch_size=p.shape[0],
                    )
            self.log(
                prefix + "loss",
                loss.detach(),
                on_step=False,
                on_epoch=True,
                prog_bar=True,
                logger=True,
                batch_size=p.shape[0],
            )
        return loss

    def forward(self, x):
        raise NotImplementedError

    def configure_optimizers(self, **kwargs):
        optimizer = torch.optim.Adamax(self.parameters(), **self.optimizer_kwargs)
        return optimizer

    @classmethod
    def test(cls, data, name, checkpoint_path):
        data.prepare_data()
        data.setup()
        name += "__" + "_".join(data.full_identifier)
        test_data = data.test_dataloader()

        # Calculate weights per class

        model = cls.load_from_checkpoint(checkpoint_path)

        trainer = pl.Trainer(
            replace_sampler_ddp=False,
        )

        test = trainer.test(model, test_data)
        print(test)

    @classmethod
    def run(
        cls,
        data,
        name,
        epochs,
        model_args: list = None,
        model_kwargs: dict = None,
        loss=torch.nn.BCELoss,
        weighted=False,
        version=None,
        **kwargs
    ):
        if model_args is None:
            model_args = []
        if model_kwargs is None:
            model_kwargs = {}
        data.prepare_data()
        data.setup()
        name += "__" + "_".join(data.full_identifier)
        train_data = data.train_dataloader()
        val_data = data.val_dataloader()

        if weighted:
            weights = model_kwargs.get("weights")
            if weights is None:
                weights = 1 + torch.sum(
                    torch.cat([data.y for data in train_data]).float(), dim=0
                )
                weights = torch.mean(weights) / weights
                name += "__weighted"
            model_kwargs["weights"] = weights
        else:
            try:
                model_kwargs.pop("weights")
            except KeyError:
                pass

        if torch.cuda.is_available():
            trainer_kwargs = dict(gpus=-1, accelerator="ddp")
        else:
            trainer_kwargs = dict(gpus=0)

        tb_logger = pl_loggers.TensorBoardLogger("logs/", name=name, version=version)
        if os.path.isdir(tb_logger.log_dir):
            raise IOError("Fixed logging directory does already exist:", tb_logger.log_dir)
        best_checkpoint_callback = ModelCheckpoint(
            dirpath=os.path.join(tb_logger.log_dir, "best_checkpoints"),
            filename="{epoch}-{val_F1Score_micro:.4f}--{val_loss:.4f}",
            save_top_k=5,
            monitor="val_loss",
            mode="min",
        )
        checkpoint_callback = ModelCheckpoint(
            dirpath=os.path.join(tb_logger.log_dir, "periodic_checkpoints"),
            filename="{epoch}-{val_F1Score_micro:.4f}--{val_loss:.4f}",
            every_n_epochs=5,
            save_top_k=-1,
            save_last=True,
            verbose=True,
        )

        # Calculate weights per class

        net = cls(*model_args, loss_cls=loss, **model_kwargs, **kwargs)

        # Early stopping seems to be bugged right now with ddp accelerator :(
        es = EarlyStopping(
            monitor="val_loss", patience=10, min_delta=0.00, verbose=False, mode="min"
        )

        trainer = pl.Trainer(
            logger=tb_logger,
            min_epochs=epochs,
            max_epochs=epochs,
            callbacks=[best_checkpoint_callback, checkpoint_callback, es],
            replace_sampler_ddp=False,
            **trainer_kwargs
        )
        trainer.fit(net, train_data, val_dataloaders=val_data)

    def pred(self, feature, batch_index=0):
        return (
            torch.sigmoid(self.predict_step(feature.to(self.device), batch_index))
            .detach()
            .cpu()
        )



import logging
import os
import pickle
import sys

from torch import nn
from torch.nn import functional as F
from torch.nn.functional import pad
import networkx as nx
import torch

from chebai.models.base import JCIBaseNet

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class ChemYK(JCIBaseNet):
    NAME = "ChemYK"

    def __init__(self, in_d, out_d, num_classes, **kwargs):
        super().__init__(num_classes, **kwargs)
        d_internal = in_d
        self.d_internal = d_internal
        self.embedding = nn.Embedding(800, d_internal)
        self.s = nn.Linear(d_internal, 1)
        self.a_l = nn.Linear(d_internal, 1)
        self.a_r = nn.Linear(d_internal, 1)
        self.w_l = nn.Linear(d_internal, d_internal)
        self.w_r = nn.Linear(d_internal, d_internal)
        self.norm = nn.LayerNorm(d_internal)
        self.output = nn.Sequential(
            nn.Linear(in_d, in_d),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(in_d, num_classes),
        )

    def forward(self, data, *args, **kwargs):
        m = self.embedding(data.x)
        max_width = m.shape[1]
        h = [m]  # torch.zeros(emb.shape[0], max_width, *emb.shape[1:])
        # h[:, 0] = emb
        for width in range(1, max_width):
            l = torch.stack(tuple(h[i][:, : (max_width - width)] for i in range(width)))
            r = torch.stack(
                tuple(h[i][:, (width - i) :] for i in range(0, width))
            ).flip(0)
            m = self.merge(l, r)
            h.append(m)
        return self.output(m).squeeze(1)

    def merge(self, l, r):
        x = torch.stack([self.a_l(l), self.a_r(r)])
        beta = torch.softmax(x, 0)
        return F.leaky_relu(
            self.attention(
                torch.sum(beta * torch.stack([self.w_l(l), self.w_r(r)]), dim=0)
            )
        )

    def attention(self, parts):
        at = torch.softmax(self.s(parts), 1)
        return torch.sum(at * parts, dim=0)



import logging
import sys

from k_gnn import avg_pool
from torch import nn
from torch_geometric import nn as tgnn
from torch_scatter import scatter_mean
import torch
import torch.nn.functional as F

from chebai.models.base import JCIBaseNet

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class JCIGraphK2Net(JCIBaseNet):
    NAME = "GNN_K2"

    def __init__(self, in_length, hidden_length, num_classes, weights=None, **kwargs):
        super().__init__(num_classes, weights, **kwargs)
        self.embedding = torch.nn.Embedding(800, in_length)

        self.conv1_1 = tgnn.GraphConv(in_length, in_length)
        self.conv1_2 = tgnn.GraphConv(in_length, in_length)
        self.conv1_3 = tgnn.GraphConv(in_length, hidden_length)

        self.conv2_1 = tgnn.GraphConv(in_length, in_length)
        self.conv2_2 = tgnn.GraphConv(in_length, in_length)
        self.conv2_3 = tgnn.GraphConv(in_length, hidden_length)

        self.output_net = nn.Sequential(
            nn.Linear(hidden_length * 2, hidden_length),
            nn.ELU(),
            nn.Linear(hidden_length, hidden_length),
            nn.ELU(),
            nn.Linear(hidden_length, num_classes),
        )

        self.dropout = nn.Dropout(0.1)

    def forward(self, x):
        a = self.embedding(x.x)
        a = self.dropout(a)

        a = F.elu(self.conv1_1(a, x.edge_index.long()))
        a = F.elu(self.conv1_2(a, x.edge_index.long()))
        a = F.elu(self.conv1_3(a, x.edge_index.long()))
        a_1 = scatter_mean(a, x.batch, dim=0)

        a = avg_pool(a, x.assignment_index_2)
        a = F.elu(self.conv2_1(a, x.edge_index.long()))
        a = F.elu(self.conv2_2(a, x.edge_index.long()))
        a = F.elu(self.conv2_3(a, x.edge_index.long()))
        a_2 = scatter_mean(a, x.batch_2, dim=0)

        a = torch.cat([a_1, a_2], dim=1)

        a = self.dropout(a)
        return self.output_net(a)



import click

from chebai import experiments
from chebai.result.base import PROCESSORS, ResultFactory
from chebai.result.molplot import AttentionOnMoleculesProcessor
from chebai.result.prediction_json import JSONResultProcessor


@click.group()
def cli():
    pass


@click.command()
@click.argument("experiment")
@click.argument("batch_size", type=click.INT)
@click.option("-g", "--group", default="default")
@click.option("--version", default=None)
@click.option("--load-prefix", default=None)
@click.option("--epochs", default=100)
@click.argument("args", nargs=-1)
def train(experiment, batch_size, group, version, load_prefix, epochs, args):
    """Run experiment identified by EXPERIMENT in batches of size BATCH_SIZE."""
    try:
        ex = experiments.EXPERIMENTS[experiment](batch_size, group, version=version)
    except KeyError:
        raise Exception(
            "Experiment ID not found. The following are available:"
            + ", ".join(experiments.EXPERIMENTS.keys())
        )
    ex.train(batch_size, epochs, *args, load_prefix=load_prefix)


@click.command()
@click.argument("experiment")
@click.argument("batch_size", type=click.INT)
@click.argument("ckpt_path")
@click.argument("args", nargs=-1)
def test(experiment, batch_size, ckpt_path, args):
    """Run experiment identified by EXPERIMENT in batches of size BATCH_SIZE."""
    try:
        ex = experiments.EXPERIMENTS[experiment](
            batch_size,
        )
    except KeyError:
        raise Exception(
            "Experiment ID not found. The following are available:"
            + ", ".join(experiments.EXPERIMENTS.keys())
        )
    ex.test(ckpt_path, *args)


@click.command()
@click.argument("experiment")
@click.argument("ckpt_path")
@click.argument("data_path")
@click.option("--processors", "-p", default=["json"], multiple=True)
@click.option("--compiled-data", "-c", is_flag=True, default=False)
@click.option("--headers", default=None)
def predict(experiment, ckpt_path, data_path, processors, compiled_data, headers):
    """Run experiment identified by EXPERIMENT in batches of size BATCH_SIZE."""
    try:
        ex = experiments.EXPERIMENTS[experiment](1)
    except KeyError:
        raise Exception(
            "Experiment ID not found. The following are available:"
            + ", ".join(experiments.EXPERIMENTS.keys())
        )
    if headers is not None:
        headers = headers.split(",")
    processor_list = []
    for p in processors:
        try:
            processor_list.append(PROCESSORS[p]())
        except KeyError:
            raise Exception(
                f"Processor {p} not found. Available processors are {', '.join(PROCESSORS.keys())}"
            )
    ex.predict(
        data_path, ckpt_path, processor_list, raw=not compiled_data, headers=headers
    )


cli.add_command(train)
cli.add_command(predict)
cli.add_command(test)



import json
from sklearn.metrics import confusion_matrix
import numpy as np
from chebai.preprocessing.datasets.tox21 import Tox21Base
import os

path = "/home/glauer/dev/ChEBI_RvNN/.experiments/electra/tox21_without_sempre/"
with open(os.path.join(path, "predictions.json"), "r") as fin:
    results = json.load(fin)
preds, labs = zip(*((row["prediction"], row["labels"]) for row in results))
preds = np.array(preds)
labs = np.array(labs)

lines = []
for dim in range(len(preds[0])):
    lines.append(f"# {Tox21Base.HEADERS[dim]}")
    has_value = labs[:, dim] == None
    p = preds[:, dim][~has_value] > 0.3
    l = labs[:, dim][~has_value]

    cm = confusion_matrix(l.astype(int), p.astype(int))
    lines.append("| | Target: False | Target: True |")
    lines.append(f"| --- | --- | --- |")
    lines.append(f"| Pred: False | {cm[0,0]} | {cm[0,1]} |")
    lines.append(f"| Pred: True | {cm[1,0]} | {cm[1,1]} |")
    lines.append("")

    precision = cm[1, 1] / (cm[1, 1] + cm[0, 1])
    recall = cm[1, 1] / (cm[1, 1] + cm[1, 0])
    lines.append(f"* Precision: {precision}")
    lines.append(f"* Recall: {recall}")
    lines.append("")
    lines.append("")
with open(os.path.join(path, "class-wise.md"), "wt") as fout:
    fout.writelines(x + "\n" for x in lines)



from tempfile import TemporaryDirectory
import logging
import random

from torch import nn
from torch.nn.functional import one_hot
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from transformers import (
    RobertaConfig,
    RobertaForMaskedLM,
    RobertaModel,
    RobertaTokenizer,
)
import torch

from chebai.models.base import JCIBaseNet

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)
MAX_LEN = 1800


class ChembertaPre(JCIBaseNet):
    NAME = "ChembertaPre"

    def __init__(self, p=0.2, **kwargs):
        super().__init__(**kwargs)
        self._p = p
        self.config = RobertaConfig(**kwargs["config"])
        self.model = RobertaForMaskedLM(self.config)

    def _get_data_and_labels(self, batch, batch_idx):
        masked = (
            torch.rand([batch.x.shape[0]], device=self.device)
            * torch.tensor(batch.lens, device=self.device)
        ).long()
        labels = one_hot(
            torch.gather(batch.x, 1, masked.unsqueeze(-1)).squeeze(-1),
            self.config.vocab_size,
        )
        features = 1 + batch.x
        features = features * (1 - one_hot(masked, batch.x.shape[-1]))
        return features, labels

    def forward(self, data):
        x = self.model(data)
        return {"logits": torch.sum(x.logits, dim=1)}


class Chemberta(JCIBaseNet):
    NAME = "Chemberta"

    def __init__(self, **kwargs):
        # Remove this property in order to prevent it from being stored as a
        # hyper parameter
        pretrained_checkpoint = (
            kwargs.pop("pretrained_checkpoint")
            if "pretrained_checkpoint" in kwargs
            else None
        )
        super().__init__(**kwargs)
        self.config = RobertaConfig(
            **kwargs["config"], output_attentions=True, num_labels=self.out_dim
        )

        if pretrained_checkpoint:
            elpre = RobertaModel.load_from_checkpoint(pretrained_checkpoint)
            with TemporaryDirectory() as td:
                elpre.electra.save_pretrained(td)
                self.electra = RobertaModel.from_pretrained(td, config=self.config)
                in_d = elpre.config.hidden_size
        else:
            self.electra = RobertaModel(config=self.config)
            in_d = self.config.hidden_size

    def forward(self, data):
        electra = self.electra(data)
        return dict(logits=electra.logits, attentions=electra.attentions)



from __future__ import absolute_import, division

import logging

try:
    from rdkit import Chem
except ModuleNotFoundError:
    pass

import networkx as nx
import numpy as np
import six
import torch

logger = logging.getLogger(__name__)


class Molecule:
    max_number_of_parents = 7

    def __init__(self, smile, logp=None, contract_rings=False):
        self.smile = smile
        self.logp = logp
        # logger.info("Parsing Molecule {:},contract rings: {:}".format(smile, contract_rings))
        self.atoms = []

        m = Chem.MolFromSmiles(smile)
        # Chem.Kekulize(self.m)

        self.no_of_atoms = m.GetNumAtoms()
        self.graph = nx.Graph()

        for i in range(self.no_of_atoms):
            atom = m.GetAtomWithIdx(i)
            self.graph.add_node(
                i, attr_dict={"atom_features": Molecule.atom_features(atom)}
            )
            for neighbour in atom.GetNeighbors():
                neighbour_idx = neighbour.GetIdx()
                bond = m.GetBondBetweenAtoms(i, neighbour_idx)
                self.graph.add_edge(
                    i,
                    neighbour_idx,
                    attr_dict={"bond_features": Molecule.bond_features(bond)},
                )

        self.create_directed_graphs()
        # self.create_feature_vectors()

    def create_directed_graphs(self):
        """
        :return:
        """
        self.directed_graphs = np.empty(
            (self.no_of_atoms, self.no_of_atoms - 1, 3), dtype=int
        )

        self.dag_to_node = {}

        # parse all the atoms one by one and get directed graph to that atom
        # as the sink node
        for idx in self.graph.nodes:
            # get shortest path from the root to all the other atoms and then reverse the edges.
            path = nx.single_source_dijkstra_path(self.graph, idx)
            G = nx.DiGraph()
            for i in self.graph.nodes:
                temp = path[i]
                temp.reverse()
                nx.add_path(G, temp)
            self.dag_to_node[idx] = G
            break

    def create_feature_vectors(self):
        """
        :return:
        """
        # create a three dimesnional matrix I,
        # such that Iij is the local input vector for jth vertex in ith DAG

        length_of_bond_features = Molecule.num_bond_features()
        length_of_atom_features = Molecule.num_atom_features()

        self.local_input_vector = np.zeros(
            (self.no_of_atoms, self.no_of_atoms, Molecule.num_of_features())
        )

        for idx in range(self.no_of_atoms):
            sorted_path = self.directed_graphs[idx, :, :]

            self.local_input_vector[
                idx, idx, :length_of_atom_features
            ] = self.get_atom_features(idx)

            no_of_incoming_edges = {}
            for i in range(self.no_of_atoms - 1):
                node1 = sorted_path[i, 0]
                node2 = sorted_path[i, 1]

                self.local_input_vector[
                    idx, node1, :length_of_atom_features
                ] = self.get_atom_features(node1)

                if node2 in no_of_incoming_edges:
                    index = no_of_incoming_edges[node2]
                    no_of_incoming_edges[node2] += 1
                    if index >= Molecule.max_number_of_parents:
                        continue
                else:
                    index = 0
                    no_of_incoming_edges[node2] = 1

                start = length_of_atom_features + index * length_of_bond_features
                end = start + length_of_bond_features

                self.local_input_vector[idx, node2, start:end] = self.get_bond_features(
                    node1, node2
                )

    def get_cycle(self):
        try:
            return nx.find_cycle(self.graph)
        except:
            return []

    def collect_atom_features(self):
        self.af = {
            node_id: torch.tensor(
                self.graph.nodes[node_id]["attr_dict"]["atom_features"]
            ).float()
            for node_id in range(self.no_of_atoms)
        }

    def get_atom_features(self, node_id):
        return self.af[node_id]

    def get_bond_features(self, node1, node2):
        attrs = self.graph.get_edge_data(node1, node2)
        return attrs["bond_features"]

    @staticmethod
    def atom_features(atom):
        return np.array(
            Molecule.one_of_k_encoding_unk(
                atom.GetSymbol(),
                [
                    "C",
                    "N",
                    "O",
                    "S",
                    "F",
                    "Si",
                    "P",
                    "Cl",
                    "Br",
                    "Mg",
                    "Na",
                    "Ca",
                    "Fe",
                    "As",
                    "Al",
                    "I",
                    "B",
                    "V",
                    "K",
                    "Tl",
                    "Yb",
                    "Sb",
                    "Sn",
                    "Ag",
                    "Pd",
                    "Co",
                    "Se",
                    "Ti",
                    "Zn",
                    "H",  # H?
                    "Li",
                    "Ge",
                    "Cu",
                    "Au",
                    "Ni",
                    "Cd",
                    "In",
                    "Mn",
                    "Zr",
                    "Cr",
                    "Pt",
                    "Hg",
                    "Pb",
                    "Unknown",
                ],
            )
            + Molecule.one_of_k_encoding(atom.GetDegree(), [0, 1, 2, 3, 4, 5])
            + Molecule.one_of_k_encoding_unk(atom.GetTotalNumHs(), [0, 1, 2, 3, 4])
            + Molecule.one_of_k_encoding_unk(
                atom.GetImplicitValence(), [0, 1, 2, 3, 4, 5]
            )
            + [atom.GetIsAromatic()]
        )

    @staticmethod
    def atom_features_of_contract_rings(degree):
        return np.array(
            Molecule.one_of_k_encoding_unk(
                "Unknown",
                [
                    "C",
                    "N",
                    "O",
                    "S",
                    "F",
                    "Si",
                    "P",
                    "Cl",
                    "Br",
                    "Mg",
                    "Na",
                    "Ca",
                    "Fe",
                    "As",
                    "Al",
                    "I",
                    "B",
                    "V",
                    "K",
                    "Tl",
                    "Yb",
                    "Sb",
                    "Sn",
                    "Ag",
                    "Pd",
                    "Co",
                    "Se",
                    "Ti",
                    "Zn",
                    "H",  # H?
                    "Li",
                    "Ge",
                    "Cu",
                    "Au",
                    "Ni",
                    "Cd",
                    "In",
                    "Mn",
                    "Zr",
                    "Cr",
                    "Pt",
                    "Hg",
                    "Pb",
                    "Unknown",
                ],
            )
            + Molecule.one_of_k_encoding(degree, [0, 1, 2, 3, 4, 5])
            + Molecule.one_of_k_encoding_unk(0, [0, 1, 2, 3, 4])
            + Molecule.one_of_k_encoding_unk(0, [0, 1, 2, 3, 4, 5])
            + [0]
        )

    @staticmethod
    def bond_features_between_contract_rings():
        return np.array([1, 0, 0, 0, 0, 0])

    @staticmethod
    def bond_features(bond):
        bt = bond.GetBondType()
        return np.array(
            [
                bt == Chem.rdchem.BondType.SINGLE,
                bt == Chem.rdchem.BondType.DOUBLE,
                bt == Chem.rdchem.BondType.TRIPLE,
                bt == Chem.rdchem.BondType.AROMATIC,
                bond.GetIsConjugated(),
                bond.IsInRing(),
            ]
        )

    @staticmethod
    def num_of_features():
        return (
            Molecule.max_number_of_parents * Molecule.num_bond_features()
            + Molecule.num_atom_features()
        )

    @staticmethod
    def one_of_k_encoding(x, allowable_set):
        if x not in allowable_set:
            raise Exception(
                "input {0} not in allowable set{1}:".format(x, allowable_set)
            )
        return list(map(lambda s: x == s, allowable_set))

    @staticmethod
    def one_of_k_encoding_unk(x, allowable_set):
        """Maps inputs not in the allowable set to the last element."""
        if x not in allowable_set:
            x = allowable_set[-1]
        return list(map(lambda s: x == s, allowable_set))

    @staticmethod
    def num_atom_features():
        # Return length of feature vector using a very simple molecule.
        m = Chem.MolFromSmiles("CC")
        alist = m.GetAtoms()
        a = alist[0]
        return len(Molecule.atom_features(a))

    @staticmethod
    def num_bond_features():
        # Return length of feature vector using a very simple molecule.
        simple_mol = Chem.MolFromSmiles("CC")
        Chem.SanitizeMol(simple_mol)
        return len(Molecule.bond_features(simple_mol.GetBonds()[0]))


if __name__ == "__main__":
    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_format)
    logger = logging.getLogger(__name__)



from abc import ABC, abstractmethod
from typing import Dict, Iterable, List
import json
import os.path

from chebai import MODULE_PATH
import torch.nn

from chebai import preprocessing as prep
from chebai.models import base, chemberta, electra, graph
from chebai.preprocessing import datasets
from chebai.result.base import ResultFactory, ResultProcessor

EXPERIMENTS = dict()


class Experiment(ABC):
    MODEL = base.JCIBaseNet
    LOSS = torch.nn.BCEWithLogitsLoss

    def __init_subclass__(cls, **kwargs):
        assert (
            cls.identifier() not in EXPERIMENTS
        ), f"Identifier {cls.identifier()} is not unique."
        if cls.identifier() is not None:
            EXPERIMENTS[cls.identifier()] = cls

    def __init__(self, batch_size, *args, version=None, **kwargs):
        self.dataset = self.build_dataset(batch_size)
        self.version=version

    @classmethod
    def identifier(cls) -> str:
        raise NotImplementedError

    @abstractmethod
    def model_kwargs(self, *args) -> Dict:
        raise NotImplementedError

    @abstractmethod
    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        raise NotImplementedError

    def train(self, batch_size, epochs, *args, **kwargs):
        self.MODEL.run(
            self.dataset,
            self.MODEL.NAME,
            epochs,
            loss=self.LOSS,
            model_kwargs=self.model_kwargs(*args),
            version=self.version,
            **kwargs
        )

    def test(self, ckpt_path, *args):
        self.MODEL.test(
            self.dataset,
            self.MODEL.NAME,
            ckpt_path,
        )

    def predict(
        self, data_path, model_ckpt, processors: Iterable[ResultProcessor], **kwargs
    ):
        model = self.MODEL.load_from_checkpoint(model_ckpt)
        result_factory = ResultFactory(model, self.dataset, processors)
        result_factory.execute(data_path, **kwargs)


class ElectraPreOnSWJ(Experiment):
    MODEL = electra.ElectraPre
    LOSS = electra.ElectraPreLoss

    @classmethod
    def identifier(cls) -> str:
        return "ElectraPre+SWJ"

    def model_kwargs(self, *args) -> Dict:
        return dict(
            optimizer_kwargs=dict(lr=1e-4),
            config=dict(
                generator=dict(
                    vocab_size=1400,
                    max_position_embeddings=1800,
                    num_attention_heads=8,
                    num_hidden_layers=6,
                    type_vocab_size=1,
                ),
                discriminator=dict(
                    vocab_size=1400,
                    max_position_embeddings=1800,
                    num_attention_heads=8,
                    num_hidden_layers=6,
                    type_vocab_size=1,
                ),
            ),
        )

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.SWJChem(batch_size, k=100)


class ChembertaPreOnSWJ(Experiment):
    MODEL = chemberta.ChembertaPre

    @classmethod
    def identifier(cls) -> str:
        return "ChembertaPre+SWJ"

    def model_kwargs(self, *args) -> Dict:
        return dict(
            optimizer_kwargs=dict(lr=1e-4),
            config=dict(
                vocab_size=1400,
                max_position_embeddings=1800,
                num_attention_heads=8,
                num_hidden_layers=6,
                type_vocab_size=1,
            ),
        )

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.SWJChem(batch_size, k=100)


class ChembertaPreBPEOnSWJ(Experiment):
    MODEL = chemberta.ChembertaPre

    @classmethod
    def identifier(cls) -> str:
        return "ChembertaPreBPE+SWJ"

    def model_kwargs(self, *args) -> Dict:
        return dict(
            optimizer_kwargs=dict(lr=1e-4),
            tokenizer_path=os.path.join(MODULE_PATH, "preprocessing/bin/BPE_SWJ"),
            config=dict(
                vocab_size=4000,
                max_position_embeddings=1800,
                num_attention_heads=8,
                num_hidden_layers=6,
                type_vocab_size=1,
            ),
        )

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.SWJBPE(
            batch_size,
            reader_kwargs=dict(
                data_path=os.path.join(MODULE_PATH, "preprocessing/bin/BPE_SWJ")
            ),
            k=100,
        )


class _ElectraExperiment(Experiment):
    MODEL = electra.Electra

    @classmethod
    def identifier(cls) -> str:
        return None

    def model_kwargs(self, *args) -> Dict:
        checkpoint_path = args[0] if len(args) > 0 else None
        return dict(
            optimizer_kwargs=dict(lr=1e-4),
            pretrained_checkpoint=checkpoint_path,
            out_dim=self.dataset.label_number,
            config=dict(
                vocab_size=1400,
                max_position_embeddings=1800,
                num_attention_heads=8,
                num_hidden_layers=6,
                type_vocab_size=1,
            )
        )


class ElectraOnJCI(_ElectraExperiment):
    @classmethod
    def identifier(cls) -> str:
        return "Electra+JCI"

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.JCITokenData(batch_size)


class ElectraOnChEBI100(_ElectraExperiment):
    MODEL = electra.Electra

    @classmethod
    def identifier(cls) -> str:
        return "Electra+ChEBI100"

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.ChEBIOver100(batch_size)


class ElectraOnJCIExt(ElectraOnJCI):
    MODEL = electra.Electra

    @classmethod
    def identifier(cls) -> str:
        return "Electra+JCIExt"

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.JCIExtendedTokenData(batch_size)


class ElectraPreOnSWJSelfies(ElectraPreOnSWJ):
    MODEL = electra.ElectraPre

    @classmethod
    def identifier(cls) -> str:
        return "ElectraPre+SWJSelfies"

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.SWJSelfies(batch_size)


class ElectraOnJCIExtSelfies(ElectraOnJCIExt):
    MODEL = electra.Electra

    @classmethod
    def identifier(cls) -> str:
        return "Electra+JCIExtSelfies"

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.JCIExtSelfies(batch_size)


class ElectraLegJCI(ElectraOnJCI):
    MODEL = electra.ElectraLegacy

    @classmethod
    def identifier(cls) -> str:
        return "ElectraLeg+JCI"


class ElectraLegJCIExt(ElectraOnJCIExt):
    MODEL = electra.ElectraLegacy

    @classmethod
    def identifier(cls) -> str:
        return "ElectraLeg+JCIExt"


class ElectraOnTox21(_ElectraExperiment):
    MODEL = electra.Electra
    LOSS = torch.nn.BCEWithLogitsLoss

    @classmethod
    def identifier(cls) -> str:
        return "Electra+Tox21"

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.Tox21Chem(batch_size)

    def model_kwargs(self, *args) -> Dict:
        d = super().model_kwargs(*args)
        d["config"]["hidden_dropout_prob"] = 0.4
        d["config"]["word_dropout"] = 0.2
        d["optimizer_kwargs"]["weight_decay"] = 1e-4
        return d


class ElectraOnTox21MoleculeNet(_ElectraExperiment):
    MODEL = electra.Electra
    LOSS = torch.nn.BCEWithLogitsLoss

    @classmethod
    def identifier(cls) -> str:
        return "Electra+Tox21MN"

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.Tox21MolNetChem(batch_size)

    def model_kwargs(self, *args) -> Dict:
        d = super().model_kwargs(*args)
        d["config"]["hidden_dropout_prob"] = 0.4
        d["config"]["word_dropout"] = 0.2
        d["optimizer_kwargs"]["weight_decay"] = 1e-4
        return d


class ElectraConeOnTox21MoleculeNet(ElectraOnChEBI100):
    MODEL = electra.ConeElectra
    LOSS = electra.ConeLoss

    @classmethod
    def identifier(cls) -> str:
        return "ElectraCone+Chebi100"

class ElectraOnTox21Challenge(_ElectraExperiment):
    @classmethod
    def identifier(cls) -> str:
        return "Electra+Tox21Chal"

    def model_kwargs(self, *args) -> Dict:
        d = super().model_kwargs(*args)
        d["config"]["hidden_dropout_prob"] = 0.5
        d["config"]["word_dropout"] = 0.3
        d["optimizer_kwargs"]["weight_decay"] = 1e-4
        return d

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.Tox21ChallengeChem(batch_size)


class ElectraBPEOnJCIExt(_ElectraExperiment):
    @classmethod
    def identifier(cls) -> str:
        return "Electra+JCIExtBPE"

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.JCIExtendedBPEData(
            batch_size,
            reader_kwargs=dict(
                data_path=os.path.join(MODULE_PATH, "preprocessing/bin/BPE_SWJ")
            ),
        )


class GATOnSWJ(Experiment):
    MODEL = graph.JCIGraphAttentionNet

    @classmethod
    def identifier(cls) -> str:
        return "GAT+JCIExt"

    def model_kwargs(self, *args) -> Dict:
        return dict(
            in_length=50,
            hidden_length=100,
        )

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.JCIGraphData(batch_size)


class GATOnTox21(Experiment):
    MODEL = graph.JCIGraphAttentionNet

    @classmethod
    def identifier(cls) -> str:
        return "GAT+Tox21"

    def model_kwargs(self, *args) -> Dict:
        return dict(
            in_length=50,
            hidden_length=100,
        )

    def build_dataset(self, batch_size) -> datasets.XYBaseDataModule:
        return datasets.Tox21Graph(batch_size)



from owlready2 import get_ontology
from lnn import Model, Predicate, Variable, World, Implies, Not
import tqdm
import fastobo
import pyhornedowl


def get_name(iri: str):
    return iri.split("/")[-1]


if __name__ == "__main__":
    formulae = []

    # Load disjointness axioms
    # onto_dis = pyhornedowl.open_ontology("/data/ontologies/chebi-disjoints.owl")
    # print("Process disjointness releation")
    # formulae += [Implies(predicates[get_name(c)](x), Not(predicates[get_name(d)](x))) for _, c,d in (ax for ax in onto_dis.get_axioms() if ax[0] == "AxiomKind::SubClassOf" and isinstance(ax[-1], str))]

    model = Model()
    x = Variable("x")
    y = Variable("y")

    onto = pyhornedowl.open_ontology("/data/ontologies/chebi.owl")

    print("Process classes")
    predicates = {get_name(c): Predicate(get_name(c)) for c in onto.get_classes()}

    print("Process subsumption releation")
    formulae += [
        Implies(predicates[get_name(c)](x), predicates[get_name(d)](x))
        for _, c, d in (
            ax
            for ax in onto.get_axioms()
            if ax[0] == "AxiomKind::SubClassOf" and isinstance(ax[-1], str)
        )
    ]

    model.add_knowledge(*formulae, world=World.AXIOM)
    model.print()



import logging
import sys

from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from chebai.models.base import JCIBaseNet

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class ChemLSTM(JCIBaseNet):
    NAME = "LSTM"

    def __init__(self, in_d, out_d, num_classes, **kwargs):
        super().__init__(num_classes, **kwargs)
        self.lstm = nn.LSTM(in_d, out_d, batch_first=True)
        self.embedding = nn.Embedding(800, 100)
        self.output = nn.Sequential(
            nn.Linear(out_d, in_d),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(in_d, num_classes),
        )

    def forward(self, data):
        x = data.x
        x_lens = data.lens
        x = self.embedding(x)
        x = pack_padded_sequence(x, x_lens, batch_first=True, enforce_sorted=False)
        x = self.lstm(x)[1][0]
        # = pad_packed_sequence(x, batch_first=True)[0]
        x = self.output(x)
        return x.squeeze(0)



import logging
import sys

from torch import nn
from torch_geometric import nn as tgnn
from torch_scatter import scatter_add, scatter_max, scatter_mean
import torch
import torch.nn.functional as F

from chebai.models.base import JCIBaseNet

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class JCIGraphNet(JCIBaseNet):
    NAME = "GNN"

    def __init__(self, in_length, hidden_length, num_classes, **kwargs):
        super().__init__(num_classes, **kwargs)
        self.embedding = torch.nn.Embedding(800, in_length)

        self.conv1 = tgnn.GraphConv(in_length, in_length)
        self.conv2 = tgnn.GraphConv(in_length, in_length)
        self.conv3 = tgnn.GraphConv(in_length, hidden_length)

        self.output_net = nn.Sequential(
            nn.Linear(hidden_length, hidden_length),
            nn.ELU(),
            nn.Linear(hidden_length, hidden_length),
            nn.ELU(),
            nn.Linear(hidden_length, num_classes),
        )

        self.dropout = nn.Dropout(0.1)

    def forward(self, x):
        a = self.embedding(x.x)
        a = self.dropout(a)
        a = F.elu(self.conv1(a, x.edge_index.long()))
        a = F.elu(self.conv2(a, x.edge_index.long()))
        a = F.elu(self.conv3(a, x.edge_index.long()))
        a = self.dropout(a)
        a = scatter_add(a, x.batch, dim=0)
        return self.output_net(a)


class JCIGraphAttentionNet(JCIBaseNet):
    NAME = "AGNN"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        in_length = kwargs.get("in_length")
        hidden_length = kwargs.get("hidden_length")
        self.embedding = torch.nn.Embedding(800, in_length)
        self.edge_embedding = torch.nn.Embedding(4, in_length)
        in_length = in_length + 10
        self.conv1 = tgnn.GATConv(
            in_length, in_length, 5, concat=False, dropout=0.1, add_self_loops=True
        )
        self.conv2 = tgnn.GATConv(
            in_length, in_length, 5, concat=False, add_self_loops=True
        )
        self.conv3 = tgnn.GATConv(
            in_length, in_length, 5, concat=False, add_self_loops=True
        )
        self.conv4 = tgnn.GATConv(
            in_length, in_length, 5, concat=False, add_self_loops=True
        )
        self.conv5 = tgnn.GATConv(
            in_length, in_length, 5, concat=False, add_self_loops=True
        )
        self.output_net = nn.Sequential(
            nn.Linear(in_length, hidden_length),
            nn.LeakyReLU(),
            nn.Linear(hidden_length, hidden_length),
            nn.LeakyReLU(),
            nn.Linear(hidden_length, 500),
        )
        self.dropout = nn.Dropout(0.1)

    def forward(self, batch):
        a = self.embedding(batch.x)
        a = self.dropout(a)
        a = torch.cat([a, torch.rand((*a.shape[:-1], 10)).to(self.device)], dim=1)
        a = F.leaky_relu(self.conv1(a, batch.edge_index.long()))
        a = F.leaky_relu(self.conv2(a, batch.edge_index.long()))
        a = F.leaky_relu(self.conv3(a, batch.edge_index.long()))
        a = F.leaky_relu(self.conv4(a, batch.edge_index.long()))
        a = F.leaky_relu(self.conv5(a, batch.edge_index.long()))
        a = self.dropout(a)
        a = scatter_mean(a, batch.batch, dim=0)
        a = self.output_net(a)
        return a



from setuptools import setup

setup(
    name="chebai",
    version="0.0.0",
    packages=["chebai", "chebai.models"],
    url="",
    license="",
    author="MGlauer",
    author_email="martin.glauer@ovgu.de",
    description="",
    install_requires=[
        "certifi",
        "idna",
        "joblib",
        "networkx",
        "numpy",
        "pandas",
        "python-dateutil",
        "pytz",
        "requests",
        "scikit-learn",
        "scipy",
        "six",
        "threadpoolctl",
        "torch",
        "typing-extensions",
        "urllib3",
        "pytorch_lightning",
        "torch-scatter",
        "torch-sparse",
        "torch-cluster",
        "torch-spline-conv",
        "torch-geometric>=1.9.0",
        "transformers==4.24.0",
        "fastobo",
        "pysmiles",
        "click",
        "scikit-network",
        "svgutils",
        "matplotlib",
        "rdkit",
        "selfies"
    ],
    extras_require={"dev": ["black", "isort", "pre-commit"]},
)



from itertools import combinations
import logging
import sys

from torch import exp, nn, tensor
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
import networkx as nx
import torch

from chebai.models.base import JCIBaseNet

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class ChemYK(JCIBaseNet):
    NAME = "ChemYK"

    def __init__(self, in_d, out_d, num_classes, **kwargs):
        super().__init__(num_classes, **kwargs)
        self.embedding = nn.Embedding(800, 100)
        self.left = nn.Linear(in_d, in_d)
        self.right = nn.Linear(in_d, in_d)
        self.w_l = nn.Linear(in_d, in_d)
        self.w_r = nn.Linear(in_d, in_d)
        self.ff_rep = nn.Linear(in_d, in_d)
        self.softmax = nn.Softmax()
        self.attention_weight = nn.Linear(in_d, in_d)
        self.top_level_attention_weight = nn.Linear(in_d, in_d)
        self.output = nn.Sequential(
            nn.Linear(in_d, in_d),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(in_d, num_classes),
        )

    def forward(self, batch, max_width=5):
        result = []
        for data in batch.x:
            # Calculate embeddings
            clusters = [
                (
                    frozenset({x, y}),
                    self.merge(
                        [
                            (
                                self.embedding(data.nodes[x]["x"]),
                                self.embedding(data.nodes[y]["x"]),
                            )
                        ]
                    ),
                )
                for x, y in data.edges
            ]
            while len(clusters[0][0]) < max_width:
                new_clusters = dict()
                for (cluster_l, value_l), (cluster_r, value_r) in combinations(
                    clusters, 2
                ):
                    if len(cluster_l.union(cluster_r)) == len(cluster_l) + 1:
                        u = cluster_l.union(cluster_r)
                        new_clusters[u] = new_clusters.get(u, []) + [(value_l, value_r)]
                clusters = [(c, self.merge(pairs)) for c, pairs in new_clusters.items()]
            # Collect graph embeddings
            ge = self.top_level_merge(clusters)
            result.append(ge)
        return self.output(torch.stack(result))

    def merge(self, pairs):
        return sum(self.fold(self._pair_merge(x, y)) for x, y in pairs)

    def _pair_merge(self, x, y):
        beta = self.softmax(torch.stack([self.left(x), self.right(y)]))
        h2 = beta[0] * self.w_l(x) + beta[1] * self.w_r(y)
        return self.ff_rep(h2) + h2

    def fold(self, h):
        return exp(self.attention_weight(h)) * h

    def top_level_merge(self, clusters):
        t = torch.stack([c for (_, c) in clusters])
        sm = self.softmax(self.top_level_attention_weight(t))
        return torch.sum(t * sm, dim=0)


def graphyk(graph: nx.Graph):
    graph.nodes()



import abc
import torch
import typing
import numpy as np
import networkx as nx


FeatureType = typing.TypeVar("FeatureType")
LabelType = typing.TypeVar("LabelType")


class StrOntEx(torch.Module):
    def __init__(self, computation_graph):
        pass



import csv
import multiprocessing as mp
import os
import pickle

from model import ChEBIRecNN
from molecule import Molecule
from pytorch_lightning import loggers as pl_loggers
from sklearn.metrics import f1_score
from torch.utils import data
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn as nn

BATCH_SIZE = 100
NUM_EPOCHS = 100
LEARNING_RATE = 0.01


def eval_model(model, dataset, test_labels):
    raw_values = []
    predictions = []
    final_scores = []

    with torch.no_grad():
        for batch in dataset:
            for molecule, label in batch:
                model_outputs = model(molecule)
                prediction = [1.0 if i > 0.5 else 0.0 for i in model_outputs]
                predictions.append(prediction)
                raw_values.append(model_outputs)
                final_scores.append(f1_score(prediction, label.tolist()))

        avg_f1 = sum(final_scores) / len(final_scores)
        return raw_values, predictions, final_scores, avg_f1


def crawl_info(DAG, sink_parents):
    topological_order = [int(i[0]) for i in DAG]
    target_nodes = [int(i[1]) for i in DAG]
    sink = target_nodes[-1]
    sources = []
    parents = {}

    for i in range(len(topological_order)):
        for j in range(len(target_nodes)):
            if topological_order[i] == target_nodes[j]:
                if topological_order[i] not in parents.keys():
                    parents[topological_order[i]] = []
                parents[topological_order[i]].append(topological_order[j])

    for node in topological_order:
        if node not in parents.keys():
            sources.append(node)

    return topological_order, sources, parents, sink, sink_parents


import random


def collate(batch):
    input, labels = zip(*batch)
    return input, torch.stack(labels)


def _execute(model, loss_fn, optimizer, data, device, with_grad=True):
    train_running_loss = 0
    data_size = 0
    f1 = 0
    model.train(with_grad)
    num_batches = len(data)
    num_batch = 0
    for molecules, labels in data:
        num_batch += 1
        optimizer.zero_grad()
        prediction = model(molecules)
        loss = loss_fn(prediction, labels)
        data_size += 1
        f1 += f1_score(prediction > 0.5, labels > 0.5, average="micro")
        train_running_loss += loss.item()
        if with_grad:
            print(f"Batch {num_batch}/{num_batches}")
            loss.backward()
            optimizer.step()
    return train_running_loss / data_size, f1 / data_size


def execute_network(
    model, loss_fn, optimizer, train_data, validation_data, epochs, device
):
    model.to(device)
    model.device = device

    for name, param in model.named_parameters():
        if param.requires_grad:
            print(name)

    columns_name = [
        "epoch",
        "train_running_loss",
        "train_running_f1",
        "eval_running_loss",
        "eval_running_f1",
    ]
    with open(r"../loss_f1_training_validation.csv", "w") as f:
        writer = csv.writer(f)
        writer.writerow(columns_name)

    for epoch in range(epochs):
        train_running_loss, train_running_f1 = _execute(
            model, loss_fn, optimizer, train_data, device, with_grad=True
        )

        with torch.no_grad():
            eval_running_loss, eval_running_f1 = _execute(
                model, loss_fn, optimizer, validation_data, device, with_grad=False
            )
        print(
            f"Epoch {epoch}: loss={train_running_loss:.5f}, f1={train_running_f1:.5f}, val_loss={eval_running_loss:.5f}, val_f1={eval_running_f1:.5f}".format(
                epoch, train_running_f1
            )
        )
        fields = [
            epoch,
            train_running_loss,
            train_running_f1,
            eval_running_loss,
            eval_running_f1,
        ]
        with open(r"../loss_f1_training_validation.csv", "a") as f:
            writer = csv.writer(f)
            writer.writerow(fields)


def prepare_data(infile):
    data = pickle.load(infile)
    infile.close()

    data_frame = pd.DataFrame.from_dict(data)
    data_frame.reset_index(drop=True, inplace=True)

    data_classes = list(data_frame.columns)
    data_classes.remove("MOLECULEID")
    data_classes.remove("SMILES")

    for col in data_classes:
        data_frame[col] = data_frame[col].astype(int)

    train_data = []
    for index, row in data_frame.iterrows():
        train_data.append(
            [
                data_frame.iloc[index].values[1],
                data_frame.iloc[index].values[2:502].tolist(),
            ]
        )

    train_df = pd.DataFrame(train_data, columns=["SMILES", "LABELS"])
    return train_df


def batchify(x, y):
    data = list(zip(x, y))
    return [
        data[i * BATCH_SIZE : (i + 1) * BATCH_SIZE]
        for i in range(1 + len(data) // BATCH_SIZE)
    ]


def load_data():
    fpath = "data/full.pickle"
    if os.path.isfile(fpath):
        with open(fpath, "rb") as f:
            (
                train_dataset,
                train_actual_labels,
                validation_dataset,
                validation_actual_labels,
            ) = pickle.load(f)
    else:
        print("reading data from files!")
        train_infile = open("../data/JCI_graph/raw/train.pkl", "rb")
        test_infile = open("../data/JCI_graph/raw/test.pkl", "rb")
        validation_infile = open("../data/JCI_graph/raw/validation.pkl", "rb")

        # test_data = prepare_data(test_infile)

        print("prepare train data!")
        train_dataset = []
        train_actual_labels = []

        for index, row in prepare_data(train_infile).iterrows():
            try:
                mol = Molecule(row["SMILES"], True)

                DAGs_meta_info = mol.dag_to_node
                train_dataset.append(mol)
                train_actual_labels.append(torch.tensor(row["LABELS"]).float())
            except:
                pass

        print("prepare validation data!")
        validation_dataset = []
        validation_actual_labels = []

        for index, row in prepare_data(validation_infile).iterrows():
            try:
                mol = Molecule(row["SMILES"], True)

                DAGs_meta_info = mol.dag_to_node

                validation_dataset.append(mol)
                validation_actual_labels.append(torch.tensor(row["LABELS"]).float())
            except:
                pass

        with open(fpath, "wb") as f:
            pickle.dump(
                (
                    train_dataset,
                    train_actual_labels,
                    validation_dataset,
                    validation_actual_labels,
                ),
                f,
            )

    return (
        train_dataset,
        train_actual_labels,
        validation_dataset,
        validation_actual_labels,
    )


def move_molecule(m):
    m.collect_atom_features()
    return m


if __name__ == "__main__":
    if torch.cuda.is_available():
        accelerator = "ddp"
        trainer_kwargs = dict(gpus=-1)
    else:
        accelerator = None
        trainer_kwargs = dict()

    (
        train_dataset,
        train_actual_labels,
        validation_dataset,
        validation_actual_labels,
    ) = load_data()
    train_data = data.DataLoader(
        list(
            zip(
                map(move_molecule, train_dataset),
                [l.float() for l in train_actual_labels],
            )
        ),
        batch_size=BATCH_SIZE,
        shuffle=True,
        collate_fn=collate,
    )
    validation_data = data.DataLoader(
        list(
            zip(
                map(move_molecule, validation_dataset),
                [l.float() for l in validation_actual_labels],
            )
        ),
        batch_size=BATCH_SIZE,
        collate_fn=collate,
    )

    model = ChEBIRecNN()

    tb_logger = pl_loggers.CSVLogger("../logs/")
    trainer = pl.Trainer(
        logger=tb_logger,
        accelerator=accelerator,
        max_epochs=NUM_EPOCHS,
        **trainer_kwargs,
    )
    trainer.fit(model, train_data, val_dataloaders=validation_data)

"""
model = ChEBIRecNN()
loss_fn = nn.BCEWithLogitsLoss()
optimizer = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE)

print('num of parameters of the model: ', sum(p.numel() for p in model.parameters() if p.requires_grad))

params = {
    'model':model,
    'loss_fn':loss_fn,
    'optimizer':optimizer,
    'train_data': train_data,
    'validation_data': validation_data,
    'epochs':NUM_EPOCHS,
    'device': device
}

print('start training!')
execute_network(**params)

torch.save(model.state_dict(), 'ChEBI_RvNN_{}_epochs'.format(NUM_EPOCHS))
print('model saved!')
"""



import random
from tempfile import TemporaryDirectory
import logging
import random
from math import pi
from torch import nn
from torch.nn.utils.rnn import (
    pack_padded_sequence,
    pad_packed_sequence,
    pad_sequence,
)
from transformers import (
    ElectraConfig,
    ElectraForMaskedLM,
    ElectraForMultipleChoice,
    ElectraForPreTraining,
    ElectraForSequenceClassification,
    ElectraModel,
    PretrainedConfig,
)
from chebai.preprocessing.reader import MASK_TOKEN_INDEX, CLS_TOKEN
import torch

from chebai.models.base import JCIBaseNet

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)


class ElectraPre(JCIBaseNet):
    NAME = "ElectraPre"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        config = kwargs["config"]
        self.generator_config = ElectraConfig(**config["generator"])
        self.generator = ElectraForMaskedLM(self.generator_config)
        self.discriminator_config = ElectraConfig(**config["discriminator"])
        self.discriminator = ElectraForPreTraining(self.discriminator_config)
        self.replace_p = 0.1

    @property
    def as_pretrained(self):
        return self.discriminator

    def _get_data_and_labels(self, batch, batch_idx):

        return dict(features=batch.x, labels=None, mask=batch.mask)

    def forward(self, data):
        features = data["features"]
        self.batch_size = batch_size = features.shape[0]
        max_seq_len = features.shape[1]

        mask = data["mask"]
        with torch.no_grad():
            dis_tar = (
                torch.rand((batch_size,), device=self.device) * torch.sum(mask, dim=-1)
            ).int()
            disc_tar_one_hot = torch.eq(
                torch.arange(max_seq_len, device=self.device)[None, :], dis_tar[:, None]
            )
            gen_tar = features[disc_tar_one_hot]
            gen_tar_one_hot = torch.eq(
                torch.arange(self.generator_config.vocab_size, device=self.device)[
                    None, :
                ],
                gen_tar[:, None],
            )

        raw_gen_out = torch.mean(
            self.generator(
                (features * ~disc_tar_one_hot) + MASK_TOKEN_INDEX * disc_tar_one_hot,
                attention_mask=mask,
            ).logits,
            dim=1,
        )

        with torch.no_grad():
            gen_best_guess = raw_gen_out.argmax(dim=-1)
            correct_mask = features[disc_tar_one_hot] == gen_best_guess
            random_tokens = torch.randint(
                self.generator_config.vocab_size, (batch_size,), device=self.device
            )
            replacements = gen_best_guess * ~correct_mask + random_tokens * correct_mask

        disc_out = self.discriminator(
            features * ~disc_tar_one_hot + replacements[:, None] * disc_tar_one_hot,
            attention_mask=mask,
        ).logits
        return (raw_gen_out, disc_out), (gen_tar_one_hot, disc_tar_one_hot)

    def _get_prediction_and_labels(self, batch, labels, output):
        return torch.softmax(output[0][1], dim=-1), output[1][1].int()

    def _get_data_for_loss(self, model_output, labels):
        return dict(input=model_output, target=None)


class ElectraPreLoss:
    def __init__(self):
        self.ce = torch.nn.CrossEntropyLoss()

    def __call__(self, target, input):
        t, p = input
        gen_pred, disc_pred = t
        gen_tar, disc_tar = p
        gen_loss = self.ce(target=torch.argmax(gen_tar.int(), dim=-1), input=gen_pred)
        disc_loss = self.ce(
            target=torch.argmax(disc_tar.int(), dim=-1), input=disc_pred
        )
        return gen_loss + disc_loss


def filter_dict(d, filter_key):
    return {str(k)[len(filter_key):]: v for k, v in
                  d.items() if
                  str(k).startswith(filter_key)}


class Electra(JCIBaseNet):
    NAME = "Electra"

    def _get_data_and_labels(self, batch, batch_idx):
        mask = pad_sequence(
            [torch.ones(l + 1, device=self.device) for l in batch.lens],
            batch_first=True,
        )
        cls_tokens = (
            torch.ones(batch.x.shape[0], dtype=torch.int, device=self.device).unsqueeze(
                -1
            )
            * CLS_TOKEN
        )
        return dict(
            features=torch.cat((cls_tokens, batch.x), dim=1),
            labels=batch.y,
            model_kwargs=dict(attention_mask=mask),
            target_mask=batch.target_mask,
        )

    @property
    def as_pretrained(self):
        return self.electra.electra

    def __init__(self, **kwargs):
        # Remove this property in order to prevent it from being stored as a
        # hyper parameter
        pretrained_checkpoint = (
            kwargs.pop("pretrained_checkpoint")
            if "pretrained_checkpoint" in kwargs
            else None
        )

        super().__init__(**kwargs)
        if not "num_labels" in kwargs["config"] and self.out_dim is not None:
            kwargs["config"]["num_labels"] = self.out_dim
        self.config = ElectraConfig(**kwargs["config"], output_attentions=True)
        self.word_dropout = nn.Dropout(kwargs["config"].get("word_dropout", 0))
        model_prefix = kwargs.get("load_prefix", None)

        in_d = self.config.hidden_size
        self.output = nn.Sequential(
            nn.Dropout(self.config.hidden_dropout_prob),
            nn.Linear(in_d, in_d),
            nn.GELU(),
            nn.Dropout(self.config.hidden_dropout_prob),
            nn.Linear(in_d, self.config.num_labels),
        )
        if pretrained_checkpoint:
            with open(pretrained_checkpoint, "rb") as fin:
                model_dict = torch.load(fin,map_location=self.device)
                if model_prefix:
                    state_dict = filter_dict(model_dict["state_dict"], model_prefix)
                else:
                    state_dict = model_dict["state_dict"]
                self.electra = ElectraModel.from_pretrained(None, state_dict={k:v for (k,v) in state_dict.items() if k.startswith("electra.")}, config=self.config)
                self.output.load_state_dict(filter_dict(state_dict,"output."))
        else:
            self.electra = ElectraModel(config=self.config)

    def _get_data_for_loss(self, model_output, labels):
        mask = model_output.get("target_mask")
        if mask is not None:
            d = model_output["logits"] * mask - 100 * ~mask
        else:
            d = model_output["logits"]
        return dict(input=d, target=labels.float())

    def _get_prediction_and_labels(self, data, labels, model_output):
        mask = model_output.get("target_mask")
        if mask is not None:
            d = model_output["logits"] * mask - 100 * ~mask
        else:
            d = model_output["logits"]
        return torch.sigmoid(d), labels.int()

    def forward(self, data, **kwargs):
        self.batch_size = data["features"].shape[0]
        inp = self.electra.embeddings.forward(data["features"])
        inp = self.word_dropout(inp)
        electra = self.electra(inputs_embeds=inp, **kwargs)
        d = electra.last_hidden_state[:, 0, :]
        return dict(
            logits=self.output(d),
            attentions=electra.attentions,
            target_mask=data.get("target_mask"),
        )


class ElectraLegacy(JCIBaseNet):
    NAME = "ElectraLeg"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.config = ElectraConfig(**kwargs["config"], output_attentions=True)

        if "pretrained_checkpoint" in kwargs:
            elpre = ElectraPre.load_from_checkpoint(kwargs["pretrained_checkpoint"])
            with TemporaryDirectory() as td:
                elpre.electra.save_pretrained(td)
                self.electra = ElectraModel.from_pretrained(td, config=self.config)
                in_d = elpre.config.hidden_size
        else:
            self.electra = ElectraModel(config=self.config)
            in_d = self.config.hidden_size

        self.output = nn.Sequential(
            nn.Linear(in_d, in_d),
            nn.ReLU(),
            nn.Linear(in_d, in_d),
            nn.ReLU(),
            nn.Linear(in_d, in_d),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(in_d, 500),
        )

    def forward(self, data):
        electra = self.electra(data)
        d = torch.sum(electra.last_hidden_state, dim=1)
        return dict(logits=self.output(d), attentions=electra.attentions)

class ConeElectra(JCIBaseNet):
    NAME = "ConeElectra"

    def _get_data_and_labels(self, batch, batch_idx):
        mask = pad_sequence(
            [torch.ones(l + 1, device=self.device) for l in batch.lens],
            batch_first=True,
        )
        cls_tokens = (
            torch.ones(batch.x.shape[0], dtype=torch.int, device=self.device).unsqueeze(
                -1
            )
            * CLS_TOKEN
        )
        return dict(
            features=torch.cat((cls_tokens, batch.x), dim=1),
            labels=batch.y,
            model_kwargs=dict(attention_mask=mask),
            target_mask=batch.target_mask,
        )

    @property
    def as_pretrained(self):
        return self.electra.electra

    def __init__(self, cone_dimensions=20, **kwargs):
        # Remove this property in order to prevent it from being stored as a
        # hyper parameter
        pretrained_checkpoint = (
            kwargs.pop("pretrained_checkpoint")
            if "pretrained_checkpoint" in kwargs
            else None
        )

        self.cone_dimensions = cone_dimensions

        super().__init__(**kwargs)
        if not "num_labels" in kwargs["config"] and self.out_dim is not None:
            kwargs["config"]["num_labels"] = self.out_dim
        self.config = ElectraConfig(**kwargs["config"], output_attentions=True)
        self.word_dropout = nn.Dropout(kwargs["config"].get("word_dropout", 0))
        model_prefix = kwargs.get("load_prefix", None)
        if pretrained_checkpoint:
            with open(pretrained_checkpoint, "rb") as fin:
                model_dict = torch.load(fin,map_location=self.device)
                if model_prefix:
                    state_dict = {str(k)[len(model_prefix):]:v for k,v in model_dict["state_dict"].items() if str(k).startswith(model_prefix)}
                else:
                    state_dict = model_dict["state_dict"]
                self.electra = ElectraModel.from_pretrained(None, state_dict=state_dict, config=self.config)
        else:
            self.electra = ElectraModel(config=self.config)

        in_d = self.config.hidden_size

        self.line_embedding = nn.Sequential(
            nn.Dropout(self.config.hidden_dropout_prob),
            nn.Linear(in_d, in_d),
            nn.GELU(),
            nn.Dropout(self.config.hidden_dropout_prob),
            nn.Linear(in_d, self.cone_dimensions),
        )

        self.cone_axes = nn.Parameter(2*pi*torch.rand((1, self.config.num_labels, self.cone_dimensions)))
        self.cone_arcs = nn.Parameter(pi*(1-2*torch.rand((1, self.config.num_labels, self.cone_dimensions))))

    def _get_data_for_loss(self, model_output, labels):
        mask = model_output.get("target_mask")
        d = model_output["predicted_vectors"]
        return dict(input=dict(predicted_vectors=d,
                               cone_axes = self.cone_axes,
                               cone_arcs = self.cone_arcs),
                    target=labels.float())

    def _get_prediction_and_labels(self, data, labels, model_output):
        mask = model_output.get("target_mask")
        d = model_output["predicted_vectors"].unsqueeze(1)

        d = in_cone_parts(d, self.cone_axes, self.cone_arcs)

        return torch.mean(d, dim=-1), labels.int()

    def forward(self, data, **kwargs):
        self.batch_size = data["features"].shape[0]
        inp = self.electra.embeddings.forward(data["features"])
        inp = self.word_dropout(inp)
        electra = self.electra(inputs_embeds=inp, **kwargs)
        d = electra.last_hidden_state[:, 0, :]
        return dict(
            predicted_vectors=self.line_embedding(d),
            attentions=electra.attentions,
            target_mask=data.get("target_mask"),
        )

def softabs(x, eps=0.01):
    return (x**2+eps)**0.5-eps**0.5

def anglify(x):
    return torch.tanh(x)*pi

def turn(vector, angle):
    v = vector - angle
    return v - (v > pi)*2*pi + (v< -pi)*2*pi

def in_cone_parts(vectors, cone_axes, cone_arcs):

        """
        # trap between -pi and pi
        cone_ax_ang = anglify(cone_axes)
        v = anglify(vectors)

        # trap between 0 and pi
        cone_arc_ang = (torch.tanh(cone_arcs)+1)*pi/2
        theta_L = cone_ax_ang - cone_arc_ang/2
        #theta_L = theta_L - (theta_L > 2*pi) * 2 * pi + (theta_L < 0) *2*pi
        theta_R = cone_ax_ang + cone_arc_ang/2
        #theta_R = theta_R - (theta_R > 2 * pi) * 2 * pi + (theta_R < 0) * 2 * pi
        dis = (torch.abs(turn(v, theta_L)) + torch.abs(turn(v, theta_R)) - cone_arc_ang)/(2*pi-cone_arc_ang)
        return dis
        """
        a = cone_axes - cone_arcs**2
        b = cone_axes + cone_arcs**2
        bigger_than_a = torch.sigmoid(vectors-a)
        smaller_than_b = torch.sigmoid(b-vectors)
        return bigger_than_a * smaller_than_b


class ConeLoss:

    def __init__(self, center_scaling=0.1):
        self.center_scaling = center_scaling

    def negate(self, ax, arc):
        offset = pi*torch.ones_like(ax)
        offset[ax >= 0] *= -1
        return ax + offset, pi - arc

    def __call__(self, target, input):
        predicted_vectors = input["predicted_vectors"].unsqueeze(1)
        cone_axes = input["cone_axes"]
        cone_arcs = input["cone_arcs"]
        memberships =  (1-1e-6)*(in_cone_parts(predicted_vectors, cone_axes, cone_arcs))
        loss = torch.nn.functional.binary_cross_entropy(memberships, target.unsqueeze(-1).expand(-1,-1,20))
        return loss